// Copyright (C) 2010-2011 Institut de Robotica i Informatica Industrial, CSIC-UPC.
// Author 
// All rights reserved.
//
// This file is part of iri-ros-pkg
// iri-ros-pkg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 
// IMPORTANT NOTE: This code has been generated through a script from the 
// iri_ros_scripts. Please do NOT delete any comments to guarantee the correctness
// of the scripts. ROS topics can be easly add by using those scripts. Please
// refer to the IRI wiki page for more information:
// http://wikiri.upc.es/index.php/Robotics_Lab

#ifndef _ground_segmentation_alg_node_h_
#define _ground_segmentation_alg_node_h_

#include <iri_base_algorithm/iri_base_algorithm.h>
#include <iridrivers/kf_based_terrain_analysis.h>

// [publisher subscriber headers]
#include "ground_segmentation_alg.h"

// [service client headers]

// [action server client headers]

/**
 * \brief IRI ROS Specific Algorithm Class
 *
 */
class GroundSegmentationAlgNode : public algorithm_base::IriBaseAlgorithm<GroundSegmentationAlgorithm>
{
private:
  std_msgs::Header common_header_;
  bool flag_new_lidar_data_;
  ros::Publisher lidar_points_ground_segmented_publisher_;
  sensor_msgs::PointCloud2 lidar_points_ground_segmented_ros_cloud_;

  ros::Publisher roadmap_markers_publisher_;
  visualization_msgs::MarkerArray roadmap_MarkerArray_msg_;

  ros::Publisher ground_dense_reconstruction_publisher_;
  sensor_msgs::PointCloud2 ground_dense_reconstruction_ros_cloud_;
  pcl::PointCloud<pcl::PointXYZRGBNormal> ground_dense_reconstruction_pcl_cloud_;

  pcl::PointCloud<pcl::PointXYZRGBNormal> roadmap_vertices_pcl_cloud_;

  ros::Publisher roadmap_edges_publisher_;
  iri_ground_segmentation::edges_array roadmap_edges_array_;

  ros::Publisher roadmap_vertices_publisher_;
  iri_ground_segmentation::vertices_array roadmap_vertices_array_;

  ros::Publisher obstacle_points_publisher_;
  sensor_msgs::PointCloud2 obstacle_points_ros_cloud_;
  pcl::PointCloud<pcl::PointXYZ> obstacle_points_pcl_cloud_;

  ros::Publisher ground_points_publisher_;
  sensor_msgs::PointCloud2 ground_points_ros_cloud_;
  pcl::PointCloud<pcl::PointXYZ> ground_points_pcl_cloud_;

  // [subscriber attributes]
  // Variables for copying the callbacks inputs
  sensor_msgs::PointCloud2 lidar_ros_cloud_;
  pcl::PointCloud<pcl::PointXYZRGBNormal> lidar_pcl_cloud_;

  ros::Subscriber lidar_subscriber_;
  void cb_lidar(const sensor_msgs::PointCloud2::ConstPtr &msg);

  pthread_mutex_t mutex_;
  void mutex_enter(void);
  void mutex_exit(void);

  void rosPointCloudtoPCLPointCloud(const sensor_msgs::PointCloud2 &ros_point_cloud,
                                    pcl::PointCloud<pcl::PointXYZRGBNormal> &pcl_point_cloud);

  void PCLPointCloudToRosPointCloud(const pcl::PointCloud<pcl::PointXYZRGBNormal> &pcl_point_cloud,
                                    sensor_msgs::PointCloud2 &ros_point_cloud);

  void PCLPointCloudToRosPointCloud(const pcl::PointCloud<pcl::PointXYZ> &pcl_point_cloud,
                                    sensor_msgs::PointCloud2 &ros_point_cloud);

  void nodesPCLPointCloudToVerticesArray(pcl::PointCloud<pcl::PointXYZRGBNormal> &nodes_pcl_point_cloud,
                                         iri_ground_segmentation::vertices_array &vertices_array);

  void splitSegmentedCloudInOnePerClass(void);

  void publishAll(void);

  void prepareForNextIteration(void);

  void showPerformanceStatistics(void);

  void mutexAll(void);

  void unmutexAll(void);

  bool isAllRequiredDataAlreadyReceived(void);

  void storeHeaderForSync(void);

  void convertInputsToSuitableFormats(void);

  void convertOutputsToSuitableFormats(void);

  // [service attributes]

  // [client attributes]

  // [action server attributes]

  // [action client attributes]

  /**
   * \brief config variable
   *
   * This variable has all the driver parameters defined in the cfg config file.
   * Is updated everytime function config_update() is called.
   */
  Config config_;
public:
  /**
   * \brief Constructor
   *
   * This constructor initializes specific class attributes and all ROS
   * communications variables to enable message exchange.
   */
  GroundSegmentationAlgNode(void);

  /**
   * \brief Destructor
   *
   * This destructor frees all necessary dynamic memory allocated within this
   * this class.
   */
  ~GroundSegmentationAlgNode(void);

protected:
  /**
   * \brief main node thread
   *
   * This is the main thread node function. Code written here will be executed
   * in every node loop while the algorithm is on running state. Loop frequency
   * can be tuned by modifying loop_rate attribute.
   *
   * Here data related to the process loop or to ROS topics (mainly data structs
   * related to the MSG and SRV files) must be updated. ROS publisher objects
   * must publish their data in this process. ROS client servers may also
   * request data to the corresponding server topics.
   */
  void mainNodeThread(void);

  /**
   * \brief dynamic reconfigure server callback
   *
   * This method is called whenever a new configuration is received through
   * the dynamic reconfigure. The derivated generic algorithm class must
   * implement it.
   *
   * \param config an object with new configuration from all algorithm
   *               parameters defined in the config file.
   * \param level  integer referring the level in which the configuration
   *               has been changed.
   */
  void node_config_update(Config &config, uint32_t level);

  /**
   * \brief node add diagnostics
   *
   * In this abstract function additional ROS diagnostics applied to the
   * specific algorithms may be added.
   */
  void addNodeDiagnostics(void);

  // [diagnostic functions]

  // [test functions]
};

#endif
