#include "ground_segmentation_alg_node.h"

GroundSegmentationAlgNode::GroundSegmentationAlgNode(void) :
    algorithm_base::IriBaseAlgorithm<GroundSegmentationAlgorithm>()
{
  //init class attributes if necessary
  if (!this->private_node_handle_.getParam("rate", this->config_.rate))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'rate' not found");
  }
  else
    this->setRate(this->config_.rate);

  //////////////// Robot related geometric parameters
  if (!this->private_node_handle_.getParam("sensor_height", this->config_.sensor_height))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'sensor_height' not found");
  }
  else
    this->alg_.lidar_configuration_.sensor_height = config_.sensor_height;

  if (!this->private_node_handle_.getParam("robot_height", this->config_.robot_height))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'robot_height' not found");
  }
  else
    this->alg_.filtering_configuration_.robot_height = config_.robot_height;

  //////////////// Parameters affecting the exploration of the pointcloud
  if (!this->private_node_handle_.getParam("ROI_delta_x_and_y", this->config_.ROI_delta_x_and_y))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'ROI_delta_x_and_y' not found");
  }
  else
    this->alg_.filtering_configuration_.ROI_delta_x_and_y = config_.ROI_delta_x_and_y;

  if (!this->private_node_handle_.getParam("ROI_shadow_area", this->config_.ROI_shadow_area))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'ROI_shadow_area' not found");
  }
  else
    this->alg_.filtering_configuration_.ROI_shadow_area = config_.ROI_shadow_area;

  if (!this->private_node_handle_.getParam("ground_reference_search_resolution_deg",
                                           this->config_.ground_reference_search_resolution_deg))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'ground_reference_search_resolution_deg' not found");
  }
  else
    this->alg_.filtering_configuration_.ground_reference_search_resolution =
        config_.ground_reference_search_resolution_deg * M_PI / 180.0;

  if (!this->private_node_handle_.getParam("elevation_grid_resolution", this->config_.elevation_grid_resolution))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'elevation_grid_resolution' not found");
  }
  else
    this->alg_.filtering_configuration_.elevation_grid_resolution = config_.elevation_grid_resolution;

  /////////// Kalman filter noise parameters
  ///////////////// initial uncertainties
  if (!this->private_node_handle_.getParam("z_initial_std_dev", this->config_.z_initial_std_dev))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'z_initial_std_dev' not found");
  }
  else
    this->alg_.filtering_configuration_.z_initial_std_dev = config_.z_initial_std_dev;

  if (!this->private_node_handle_.getParam("initial_angular_std_dev_deg", this->config_.initial_angular_std_dev_deg))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'initial_angular_std_dev_deg' not found");
  }
  else
  {
    // we estimate the slopes in x and y axis, so it is the tangent of the angle, but we set the values in degrees because
    // it results more intuitive
    this->alg_.filtering_configuration_.roll_initial_std_dev = tan(config_.initial_angular_std_dev_deg * M_PI / 180.0);
    this->alg_.filtering_configuration_.pitch_initial_std_dev = tan(config_.initial_angular_std_dev_deg * M_PI / 180.0);
  }
  ////////////////// propagation additive noises
  if (!this->private_node_handle_.getParam("propagation_z_additive_noise_per_meter",
                                           this->config_.propagation_z_additive_noise_per_meter))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'propagation_z_additive_noise_per_meter' not found");
  }
  else
    this->alg_.filtering_configuration_.propagation_z_additive_noise = config_.propagation_z_additive_noise_per_meter;

  if (!this->private_node_handle_.getParam("propagation_additive_noise_deg_per_meter",
                                           this->config_.propagation_additive_noise_deg_per_meter))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'propagation_additive_noise_deg_per_meter' not found");
  }
  else
    this->alg_.filtering_configuration_.propagation_additive_noise = tan(
        config_.propagation_additive_noise_deg_per_meter * M_PI / 180.0);
  ///////////////// measurement noise
  if (!this->private_node_handle_.getParam("z_observation_std_dev", this->config_.z_observation_std_dev))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'z_observation_std_dev' not found");
  }
  else
    this->alg_.filtering_configuration_.z_observation_std_dev = config_.z_observation_std_dev;

  /////////////////// threshold for outlier rejection and classification
  if (!this->private_node_handle_.getParam("mahalanobis_threshold", this->config_.mahalanobis_threshold))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'mahalanobis_threshold' not found");
  }
  else
    this->alg_.filtering_configuration_.mahalanobis_threshold = config_.mahalanobis_threshold;

  /////////////////// Neural Network parameters
  if (!this->private_node_handle_.getParam("use_neural_network", this->config_.use_neural_network))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'use_neural_network' not found");
  }
  else
    this->alg_.filtering_configuration_.use_neural_network = config_.use_neural_network;

  if (!this->private_node_handle_.getParam("extract_data_to_external_training_of_the_network",
                                           this->config_.extract_data_to_external_training_of_the_network))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'extract_data_to_external_training_of_the_network' not found");
  }
  else
    this->alg_.filtering_configuration_.extract_data_to_external_training_of_the_network =
        config_.extract_data_to_external_training_of_the_network;

  if (!this->private_node_handle_.getParam("dataset_filename_with_global_path",
                                           this->config_.dataset_filename_with_global_path))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'dataset_filename_with_global_path' not found");
  }
  else
    this->alg_.filtering_configuration_.dataset_filename_with_global_path =
        config_.dataset_filename_with_global_path;

  if (!this->private_node_handle_.getParam("neural_net_filename_with_global_path",
                                           this->config_.neural_net_filename_with_global_path))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'neural_net_filename_with_global_path' not found");
  }
  else
    this->alg_.filtering_configuration_.neural_net_filename_with_global_path =
        config_.neural_net_filename_with_global_path;

  if (!this->private_node_handle_.getParam("neural_net_number_of_features",
                                           this->config_.neural_net_number_of_features))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'neural_net_number_of_features' not found");
  }
  else
    this->alg_.filtering_configuration_.neural_net_number_of_features =
        config_.neural_net_number_of_features;

  if (!this->private_node_handle_.getParam("neural_net_number_of_neurons_in_hidden_layer",
                                           this->config_.neural_net_number_of_neurons_in_hidden_layer))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'neural_net_number_of_neurons_in_hidden_layer' not found");
  }
  else
    this->alg_.filtering_configuration_.neural_net_number_of_neurons_in_hidden_layer =
        config_.neural_net_number_of_neurons_in_hidden_layer;

  if (!this->private_node_handle_.getParam("neural_net_number_of_neurons_in_output_layer",
                                           this->config_.neural_net_number_of_neurons_in_output_layer))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'neural_net_number_of_neurons_in_output_layer' not found");
  }
  else
    this->alg_.filtering_configuration_.neural_net_number_of_neurons_in_output_layer =
        config_.neural_net_number_of_neurons_in_output_layer;

  ////////////////// labeling parameters
  if (!this->private_node_handle_.getParam("max_pred_std_dev_for_labelling",
                                           this->config_.max_pred_std_dev_for_labelling))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'max_pred_std_dev_for_labelling' not found");
  }
  else
    this->alg_.filtering_configuration_.max_pred_std_dev_for_labelling = config_.max_pred_std_dev_for_labelling;

  if (!this->private_node_handle_.getParam("score_threshold", this->config_.score_threshold))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'score_threshold' not found");
  }
  else
    this->alg_.filtering_configuration_.score_threshold = config_.score_threshold;

  if (!this->private_node_handle_.getParam("classify_not_labeled_points_as_obstacles",
                                           this->config_.classify_not_labeled_points_as_obstacles))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'classify_not_labeled_points_as_obstacles' not found");
  }
  else
    this->alg_.filtering_configuration_.classify_not_labeled_points_as_obstacles =
        config_.classify_not_labeled_points_as_obstacles;

  if (!this->private_node_handle_.getParam("ground_threshold_in_not_analyzed_areas",
                                           this->config_.ground_threshold_in_not_analyzed_areas))
  {
    ROS_WARN(
        "GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'ground_threshold_in_not_analyzed_areas' not found");
  }
  else
    this->alg_.filtering_configuration_.ground_threshold_in_not_analyzed_areas =
        config_.ground_threshold_in_not_analyzed_areas;

  ////////////////// visualization and debug parameters
  if (!this->private_node_handle_.getParam("measure_performance", this->config_.measure_performance))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'measure_performance' not found");
  }
  else
    this->alg_.filtering_configuration_.measure_performance = config_.measure_performance;

  if (!this->private_node_handle_.getParam("show_dense_reconstruction", this->config_.show_dense_reconstruction))
  {
    ROS_WARN("GroundSegmentationAlgNode::GroundSegmentationAlgNode: param 'show_dense_reconstruction' not found");
  }
  else
    this->alg_.filtering_configuration_.show_dense_reconstruction = config_.show_dense_reconstruction;

  // [init subscribers]
  this->lidar_subscriber_ = this->private_node_handle_.subscribe("lidar_points_extended_structure", 1,
                                                                 &GroundSegmentationAlgNode::cb_lidar, this);

  // [init publishers]

  //////// Segmentation outputs /////////////////////
  // This pointcloud has several point classes represented with different colors
  this->lidar_points_ground_segmented_publisher_ = this->private_node_handle_.advertise < sensor_msgs::PointCloud2
      > ("lidar_points_ground_segmented", 1);

  this->obstacle_points_publisher_ = this->private_node_handle_.advertise < sensor_msgs::PointCloud2
      > ("obstacle_points", 1);

  this->ground_points_publisher_ = this->private_node_handle_.advertise < sensor_msgs::PointCloud2
      > ("ground_points", 1);

  //////// Roadmap outputs //////////////////////////
  this->roadmap_edges_publisher_ = this->private_node_handle_.advertise < iri_ground_segmentation::edges_array
      > ("roadmap_edges", 1);

  this->roadmap_vertices_publisher_ = this->private_node_handle_.advertise < iri_ground_segmentation::vertices_array
      > ("roadmap_vertices", 1);

  //////// Visualization outputs ///////////////////
  this->roadmap_markers_publisher_ = this->private_node_handle_.advertise < visualization_msgs::MarkerArray
      > ("roadmap_markers", 1);

  this->ground_dense_reconstruction_publisher_ = this->private_node_handle_.advertise < sensor_msgs::PointCloud2
      > ("ground_dense_reconstruction", 1);

  flag_new_lidar_data_ = false;
  pthread_mutex_init(&this->mutex_, NULL);
  // [init services]

  std::cout << "Ground segmentation node instantiated!!" << std::endl;
}

GroundSegmentationAlgNode::~GroundSegmentationAlgNode(void)
{
  this->showPerformanceStatistics();
  pthread_mutex_destroy(&this->mutex_);
}

void GroundSegmentationAlgNode::rosPointCloudtoPCLPointCloud(const sensor_msgs::PointCloud2 &ros_point_cloud,
                                                             pcl::PointCloud<pcl::PointXYZRGBNormal> &pcl_point_cloud)
{
  pcl::PCLPointCloud2 aux;
  pcl_conversions::toPCL(ros_point_cloud, aux);
  pcl::fromPCLPointCloud2(aux, pcl_point_cloud);
}

void GroundSegmentationAlgNode::PCLPointCloudToRosPointCloud(
    const pcl::PointCloud<pcl::PointXYZRGBNormal> &pcl_point_cloud, sensor_msgs::PointCloud2 &ros_point_cloud)
{
  pcl::PCLPointCloud2 aux;
  toPCLPointCloud2(pcl_point_cloud, aux);
  pcl_conversions::fromPCL(aux, ros_point_cloud);
  ros_point_cloud.header = common_header_;
}

void GroundSegmentationAlgNode::PCLPointCloudToRosPointCloud(const pcl::PointCloud<pcl::PointXYZ> &pcl_point_cloud,
                                                             sensor_msgs::PointCloud2 &ros_point_cloud)
{
  pcl::PCLPointCloud2 aux;
  toPCLPointCloud2(pcl_point_cloud, aux);
  pcl_conversions::fromPCL(aux, ros_point_cloud);
  ros_point_cloud.header = common_header_;
}

void GroundSegmentationAlgNode::nodesPCLPointCloudToVerticesArray(
    pcl::PointCloud<pcl::PointXYZRGBNormal> &nodes_pcl_point_cloud,
    iri_ground_segmentation::vertices_array &vertices_array)
{
  roadmap_vertices_array_.header = common_header_;
  for (pcl::PointCloud<pcl::PointXYZRGBNormal>::iterator i = nodes_pcl_point_cloud.begin();
      i != nodes_pcl_point_cloud.end(); ++i)
  {
    pcl::PointXYZRGBNormal vertex = *i;
    iri_ground_segmentation::vertex custom_msg_vertex;

    custom_msg_vertex.id = vertex.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX];

    custom_msg_vertex.x = vertex.x;
    custom_msg_vertex.y = vertex.y;

    custom_msg_vertex.z = vertex.z;
    custom_msg_vertex.x_slope = vertex.data_c[GRND_REF_DATA_C_2_PITCH];
    custom_msg_vertex.y_slope = vertex.data_c[GRND_REF_DATA_C_1_ROLL];

    custom_msg_vertex.z_variance = vertex.data_c[GRND_REF_DATA_C_3_Z_VARIANCE];
    custom_msg_vertex.x_slope_variance = vertex.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE];
    custom_msg_vertex.y_slope_variance = vertex.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE];

    roadmap_vertices_array_.vertices_array.push_back(custom_msg_vertex);
  }
}

void GroundSegmentationAlgNode::mutexAll(void)
{
  this->mutex_enter();
}

void GroundSegmentationAlgNode::unmutexAll(void)
{
  this->mutex_exit();
}

bool GroundSegmentationAlgNode::isAllRequiredDataAlreadyReceived(void)
{
  bool ready_to_process = false;
  this->mutex_enter();
  if (flag_new_lidar_data_)
  {
    ready_to_process = true;
    //std::cout << "GroundSegmentationAlgNode::isAllRequiredDataAlreadyReceived..." << std::endl;
    //std::cout << "Ready_to_process = " << ready_to_process << std::endl;
  }
  this->mutex_exit();
  return (ready_to_process);
}

void GroundSegmentationAlgNode::storeHeaderForSync(void)
{
  common_header_ = lidar_ros_cloud_.header;
}

void GroundSegmentationAlgNode::convertInputsToSuitableFormats(void)
{
  //std::cout << "GroundSegmentationAlgNode::convertInputsToSuitableFormats..." << std::endl;
  this->rosPointCloudtoPCLPointCloud(lidar_ros_cloud_, lidar_pcl_cloud_);
//  std::cout << "x = " << lidar_pcl_cloud_.points[100].x << std::endl;
//  std::cout << "y = " << lidar_pcl_cloud_.points[100].y << std::endl;
//  std::cout << "z = " << lidar_pcl_cloud_.points[100].z << std::endl;
//
//  std::cout << "r = " << (int)lidar_pcl_cloud_.points[100].r << std::endl;
//  std::cout << "g = " << (int)lidar_pcl_cloud_.points[100].g << std::endl;
//  std::cout << "b = " << (int)lidar_pcl_cloud_.points[100].b << std::endl;
//
//  std::cout << "DATA_N_0_INTENSITY = " << lidar_pcl_cloud_.points[100].data_n[DATA_N_0_INTENSITY] << std::endl;
//  std::cout << "DATA_N_1_INTENSITY_VARIANCE = " << lidar_pcl_cloud_.points[100].data_n[DATA_N_1_INTENSITY_VARIANCE] << std::endl;
//  std::cout << "DATA_N_2_Z_MEAN = " << lidar_pcl_cloud_.points[100].data_n[DATA_N_2_Z_MEAN] << std::endl;
//  std::cout << "DATA_N_3_Z_VARIANCE = " << lidar_pcl_cloud_.points[100].data_n[DATA_N_3_Z_VARIANCE] << std::endl;
//
//  std::cout << "DATA_C_0_RGB_CAST_INTO_FLOAT = " << lidar_pcl_cloud_.points[100].data_c[DATA_C_0_RGB_CAST_INTO_FLOAT] << std::endl;
//  std::cout << "DATA_C_1_ID_CLASS = " << lidar_pcl_cloud_.points[100].data_c[DATA_C_1_ID_CLASS] << std::endl;
//  std::cout << "DATA_C_2_MEAN_INTENSITY = " << lidar_pcl_cloud_.points[100].data_c[DATA_C_2_MEAN_INTENSITY] << std::endl;
//  std::cout << "DATA_C_3_ORIGINAL_INDEX = " << lidar_pcl_cloud_.points[100].data_c[DATA_C_3_ORIGINAL_INDEX] << std::endl;

  flag_new_lidar_data_ = false;
}

void GroundSegmentationAlgNode::splitSegmentedCloudInOnePerClass(void)
{
  for (pcl::PointCloud<pcl::PointXYZRGBNormal>::iterator i = lidar_pcl_cloud_.begin(); i != lidar_pcl_cloud_.end(); ++i)
  {
    pcl::PointXYZRGBNormal classified_point = *i;
    pcl::PointXYZ regular_point;

    regular_point.x = classified_point.x;
    regular_point.y = classified_point.y;
    regular_point.z = classified_point.z;

    switch ((int)std::floor(classified_point.data_c[DATA_C_1_ID_CLASS]))
    {
      case CLASS_GROUND:
        ground_points_pcl_cloud_.push_back(regular_point);
        break;
      case CLASS_OBSTACLE:
        obstacle_points_pcl_cloud_.push_back(regular_point);
        break;
      case CLASS_OVERHANGING_OBSTACLE:
        // If we want to create a pointcloud with the overhanging obstacles, we should populate it here
        break;
      case CLASS_UNKNOWN:
        // If we ever want to create a pointcloud with unknown class points, we should populate it here
        break;
      default:
        //std::cout << "GroundSegmentationAlgNode::splitSegmentedCloudInOnePerClass -> unexpected class value of: "
        //    << classified_point.data_c[DATA_C_1_ID_CLASS] << std::endl;
        break;
    }
  }
}

void GroundSegmentationAlgNode::convertOutputsToSuitableFormats(void)
{
  //std::cout << "Extracted header = " << std::endl << common_header_;
  //std::cout << "Original  header = " << std::endl << lidar_ros_cloud_.header << std::endl;

  this->PCLPointCloudToRosPointCloud(lidar_pcl_cloud_, lidar_points_ground_segmented_ros_cloud_);

  this->PCLPointCloudToRosPointCloud(ground_dense_reconstruction_pcl_cloud_, ground_dense_reconstruction_ros_cloud_);

  this->nodesPCLPointCloudToVerticesArray(roadmap_vertices_pcl_cloud_, roadmap_vertices_array_);

  this->splitSegmentedCloudInOnePerClass();

  this->PCLPointCloudToRosPointCloud(obstacle_points_pcl_cloud_, obstacle_points_ros_cloud_);

  this->PCLPointCloudToRosPointCloud(ground_points_pcl_cloud_, ground_points_ros_cloud_);
}

void GroundSegmentationAlgNode::publishAll(void)
{
  //std::cout << "Starting GroundSegmentationAlgNode::publishAll.." << std::endl;

  // Segmentation topics
  this->lidar_points_ground_segmented_publisher_.publish(this->lidar_points_ground_segmented_ros_cloud_);
  this->obstacle_points_publisher_.publish(this->obstacle_points_ros_cloud_);
  this->ground_points_publisher_.publish(this->ground_points_ros_cloud_);

  // Roadmap topics
  this->roadmap_edges_publisher_.publish(this->roadmap_edges_array_);
  this->roadmap_vertices_publisher_.publish(this->roadmap_vertices_array_);

  // Visualization topics
  this->ground_dense_reconstruction_publisher_.publish(this->ground_dense_reconstruction_ros_cloud_);


  visualization_msgs::Marker markerD;
  markerD.header = common_header_;
  markerD.action = visualization_msgs::Marker::DELETEALL;
  this->roadmap_markers_publisher_.publish(markerD);

  this->roadmap_markers_publisher_.publish(this->roadmap_MarkerArray_msg_);
}

void GroundSegmentationAlgNode::prepareForNextIteration(void)
{
  // Segmentation
  lidar_pcl_cloud_.clear();
  obstacle_points_pcl_cloud_.clear();
  ground_points_pcl_cloud_.clear();

  // Roadmap
  roadmap_edges_array_.edges_array.clear();
  roadmap_vertices_array_.vertices_array.clear();

  // Visualization
  ground_dense_reconstruction_pcl_cloud_.clear();
  roadmap_MarkerArray_msg_.markers.clear();

  // Aux pointcloud to store vertices information (internally vertices are organized in a pointcloud)
  roadmap_vertices_pcl_cloud_.clear();
}

void GroundSegmentationAlgNode::showPerformanceStatistics(void)
{
  std::cout << "--------------------- Final Performance statistics: ---------------------------------" << std::endl;
  std::cout << "--------------------- ALGORITHM: ---------------------------------" << std::endl;
  this->alg_.showPerformanceStatistics();
  std::cout << "-------------------------------------------------------------------------------" << std::endl;
  std::cout << "--------------------- PROCESSOR: ---------------------------------" << std::endl;
  this->alg_.processor_.showPerformanceStatistics();
  std::cout << "-------------------------------------------------------------------------------" << std::endl;
}

void GroundSegmentationAlgNode::mainNodeThread(void)
{
  if (this->isAllRequiredDataAlreadyReceived())
  {
    this->mutex_enter(); // We pause the callback until we have copied all the input data

    this->storeHeaderForSync(); // Header
    this->convertInputsToSuitableFormats(); // and point cloud

    this->mutex_exit(); // Then we can free the callback

    this->alg_.lock();
    //std::cout << "Starting ground segmentation!" << std::endl;
    this->alg_.segmentGround(common_header_, lidar_pcl_cloud_, roadmap_vertices_pcl_cloud_, roadmap_edges_array_,
                             ground_dense_reconstruction_pcl_cloud_, roadmap_MarkerArray_msg_);
    //std::cout << "Ground segmentation finished!" << std::endl;
    this->alg_.unlock();

    this->convertOutputsToSuitableFormats();
    this->publishAll();
    this->prepareForNextIteration();
  }
}

/*  [subscriber callbacks] */

void GroundSegmentationAlgNode::cb_lidar(const sensor_msgs::PointCloud2::ConstPtr &msg)
{
  this->mutex_enter();
  lidar_ros_cloud_ = *msg;
  //std::cout << "GroundSegmentationAlgNode::cb_lidar --> lidar msg received!" << std::endl;
  assert(msg != NULL && "Null pointer!!! in function cb_lidar!");

  flag_new_lidar_data_ = true;

  this->mutex_exit();
}

void GroundSegmentationAlgNode::mutex_enter(void)
{
  pthread_mutex_lock(&this->mutex_);
}

void GroundSegmentationAlgNode::mutex_exit(void)
{
  pthread_mutex_unlock(&this->mutex_);
}

/*  [service callbacks] */

/*  [action callbacks] */

/*  [action requests] */

void GroundSegmentationAlgNode::node_config_update(Config &config, uint32_t level)
{
  this->alg_.lock();
  this->config_ = config;

  if (config_.rate != this->getRate())
    this->setRate(config_.rate);

  // Robot related geometric parameters
  this->alg_.lidar_configuration_.sensor_height = config.sensor_height;
  this->alg_.filtering_configuration_.robot_height = config.robot_height;

  // Parameters affecting the exploration of the pointcloud
  this->alg_.filtering_configuration_.ROI_delta_x_and_y = config.ROI_delta_x_and_y;
  this->alg_.filtering_configuration_.ROI_shadow_area = config.ROI_shadow_area;
  this->alg_.filtering_configuration_.ground_reference_search_resolution = config.ground_reference_search_resolution_deg
      * M_PI / 180.0;
  this->alg_.filtering_configuration_.elevation_grid_resolution = config.elevation_grid_resolution;

  // Kalman filter noise parameters
  // Initial uncertainties
  this->alg_.filtering_configuration_.z_initial_std_dev = config.z_initial_std_dev;
  this->alg_.filtering_configuration_.roll_initial_std_dev = tan(config.initial_angular_std_dev_deg * M_PI / 180.0);
  this->alg_.filtering_configuration_.pitch_initial_std_dev = tan(config.initial_angular_std_dev_deg * M_PI / 180.0);
  // Propagation additive noises
  this->alg_.filtering_configuration_.propagation_z_additive_noise = config.propagation_z_additive_noise_per_meter;
  this->alg_.filtering_configuration_.propagation_additive_noise = tan(
      config.propagation_additive_noise_deg_per_meter * M_PI / 180.0);
  // Measurement noise
  this->alg_.filtering_configuration_.z_observation_std_dev = config.z_observation_std_dev;

  // threshold for outlier rejection and classification
  this->alg_.filtering_configuration_.mahalanobis_threshold = config.mahalanobis_threshold;

  // Neural Network parameters
  this->alg_.filtering_configuration_.use_neural_network = config.use_neural_network;
  this->alg_.filtering_configuration_.extract_data_to_external_training_of_the_network =
      config.extract_data_to_external_training_of_the_network;

  // labeling parameters
  this->alg_.filtering_configuration_.score_threshold = config_.score_threshold;
  this->alg_.filtering_configuration_.max_pred_std_dev_for_labelling = config.max_pred_std_dev_for_labelling;
  this->alg_.filtering_configuration_.classify_not_labeled_points_as_obstacles =
      config_.classify_not_labeled_points_as_obstacles;

  // visualization and debug parameters
  this->alg_.filtering_configuration_.measure_performance = config.measure_performance;
  this->alg_.filtering_configuration_.show_dense_reconstruction = config.show_dense_reconstruction;

  std::cout << "GroundSegmentationAlgNode Configured! --> values received: " << std::endl;

  std::cout << "rate set to: " << config_.rate << " Hz" << std::endl;

  std::cout << "sensor_height               =                   " << this->alg_.lidar_configuration_.sensor_height
      << std::endl;
  std::cout << "robot_height   = " << this->alg_.filtering_configuration_.robot_height << std::endl;

  std::cout << "ROI_delta_x_and_y =                   " << this->alg_.filtering_configuration_.ROI_delta_x_and_y
      << std::endl;

  std::cout << "ROI_shadow_area =                   " << this->alg_.filtering_configuration_.ROI_shadow_area
      << std::endl;

  std::cout << "ground_reference_search_resolution (deg) =            "
      << this->alg_.filtering_configuration_.ground_reference_search_resolution * 180.0 / M_PI << std::endl;
  std::cout << "elevation_grid_resolution   =                   "
      << this->alg_.filtering_configuration_.elevation_grid_resolution << std::endl;

  std::cout << "z_initial_std_dev           =                   "
      << this->alg_.filtering_configuration_.z_initial_std_dev << std::endl;
  std::cout << "roll_initial_std_dev (deg)  =                   "
      << atan((double)this->alg_.filtering_configuration_.roll_initial_std_dev) * 180.0 / M_PI << std::endl;
  std::cout << "pitch_initial_std_dev (deg) =                   "
      << atan((double)this->alg_.filtering_configuration_.pitch_initial_std_dev) * 180.0 / M_PI << std::endl;
  std::cout << "propagation_z_additive_noise in m/m  =          "
      << this->alg_.filtering_configuration_.propagation_z_additive_noise << std::endl;
  std::cout << "propagation_additive_noise in deg/m  =          "
      << atan((double)this->alg_.filtering_configuration_.propagation_additive_noise) * 180.0 / M_PI << std::endl;
  std::cout << "z_observation_std_dev       =                   "
      << this->alg_.filtering_configuration_.z_observation_std_dev << std::endl;

  std::cout << "mahalanobis_threshold       =                   "
      << this->alg_.filtering_configuration_.mahalanobis_threshold << std::endl;

  std::cout << "score_threshold       =                   "
      << this->alg_.filtering_configuration_.score_threshold << std::endl;

  std::cout << "use_neural_network       =                   "
      << this->alg_.filtering_configuration_.use_neural_network << std::endl;

  std::cout << "extract_data_to_external_training_of_the_network       =                   "
      << this->alg_.filtering_configuration_.extract_data_to_external_training_of_the_network << std::endl;

  std::cout << "max_pred_std_dev_for_labelling      =                   "
      << this->alg_.filtering_configuration_.max_pred_std_dev_for_labelling << std::endl;
  std::cout << "classify_not_labeled_points_as_obstacles =  "
      << this->alg_.filtering_configuration_.classify_not_labeled_points_as_obstacles << std::endl;

  std::cout << "measure_performance  = " << this->alg_.filtering_configuration_.measure_performance << std::endl;
  std::cout << "show_dense_reconstruction  = " << this->alg_.filtering_configuration_.show_dense_reconstruction
      << std::endl;

  this->alg_.unlock();
}

void GroundSegmentationAlgNode::addNodeDiagnostics(void)
{
}

/* main function */
int main(int argc, char *argv[])
{
  return algorithm_base::main < GroundSegmentationAlgNode
      > (argc, argv, "two_segmented_inputs_vehicle_detector_alg_node");
}
